@@ -30,6 +30,9 @@ class LiveClient(WebSocketApp):
         self.logger = self.parent_app.logger
         self.disconnect_flag = False
         self.heartbeat_check_thread = Thread(target=self.check_heartbeat, daemon=True)
+        self.trade_callback = None
+        self.bar_callback = None
+        self.bidask_callback = None
         # if self.parent_app.live_port == 8086 or self.parent_app.live_port == 8084:
         #     # self.heartbeat_check_thread.start()
         #     pass
@@ -69,7 +72,8 @@ class LiveClient(WebSocketApp):
         return time_series[start_gap_index], time_series[start_gap_index+1]
 
     def recover_from_time_missed(self, start_time, end_time):
-        self.logger.info(f"{Style.BRIGHT}{Fore.YELLOW}Initiating recovery from {Fore.GREEN}{start_time}{Fore.YELLOW} till {Fore.GREEN}{end_time}{Fore.YELLOW} which are last green heartbeats from server...{Style.RESET_ALL}")
+        self.logger.info(f"{Style.BRIGHT}{Fore.YELLOW}Initiating recovery from {Fore.GREEN}{start_time}{Fore.YELLOW} till {Fore.GREEN}{end_time}{Fore.YELLOW} "
+                         f"which are last green heartbeats from server...{Style.RESET_ALL}")
 
     def reconnect(self):
         self.close()
@@ -103,8 +107,10 @@ class LiveClient(WebSocketApp):
             if msg['message'] == 'HeartBeat':
                 self.handle_heartbeat(msg['timestamp'])
             elif msg['message'] == 'TrueData Real Time Data Service':  # Connection success message
-                # self.logger.info(f"{Style.BRIGHT}{Fore.WHITE}You have subscribed for {msg['maxsymbols']} symbols across {msg['segments']} until {msg['validity']} with type of stream as {msg['subscription']}...{Style.RESET_ALL}")
-                print(f"{Style.NORMAL}{Fore.BLUE}Connected successfully to {msg['message']}... {Style.RESET_ALL}")
+                # self.logger.info(f"{Style.BRIGHT}{Fore.WHITE}You have subscribed for {msg['maxsymbols']} "
+                #                  f"symbols across {msg['segments']} until {msg['validity']} with type of stream "
+                #                  f"as {msg['subscription']}...{Style.RESET_ALL}")
+                self.logger.warning(f"{Style.NORMAL}{Fore.BLUE}Connected successfully to {msg['message']}... {Style.RESET_ALL}")
                 self.subscription_type = msg['subscription']
             elif msg['message'] in ['symbols added', 'touchline']:
                 self.add_contract_details(msg['symbollist'])
@@ -125,16 +131,15 @@ class LiveClient(WebSocketApp):
             for req_id in self.parent_app.symbol_mkt_id_map[contract.upper()]:
                 del self.parent_app.live_data[req_id]
                 del self.parent_app.touchline_data[req_id]
-                self.parent_app.symbol_mkt_id_map[contract.upper()].remove(req_id)
                 if '+' in self.subscription_type:
                     del self.parent_app.min_live_data[req_id]
+            del self.parent_app.symbol_mkt_id_map[contract.upper()]
 
     def add_contract_details(self, contracts_list):
         for contract_details in contracts_list:
             if contract_details is not None:
                 self.contract_mapping[int(contract_details[1])] = symbol = contract_details[0]
                 for ticker_id in self.parent_app.symbol_mkt_id_map[symbol]:
-                    # print(ticker_id)
                     self.parent_app.touchline_data[ticker_id].symbol_id = int(contract_details[1])
                     self.parent_app.touchline_data[ticker_id].timestamp = datetime.strptime(contract_details[2], '%Y-%m-%dT%H:%M:%S')
                     self.parent_app.touchline_data[ticker_id].symbol = symbol
@@ -156,8 +161,7 @@ class LiveClient(WebSocketApp):
                     self.parent_app.live_data[ticker_id].populate_using_touchline(self.parent_app.live_data[ticker_id], self.parent_app.touchline_data[ticker_id])
                     if '+' in self.subscription_type:
                         self.parent_app.min_live_data[ticker_id].populate_using_touchline(self.parent_app.min_live_data[ticker_id], self.parent_app.touchline_data[ticker_id])
-
-                # print(f"Updated touchline data for {symbol} and related live data objects...") #  self.logger.info
+                self.logger.info(f"Updated touchline data for {symbol} and related live data objects...")
             else:
                 self.logger.debug(f'{Style.BRIGHT}{Fore.YELLOW}Probable repeated symbol...{Style.RESET_ALL}')
 
@@ -165,15 +169,14 @@ class LiveClient(WebSocketApp):
         try:
             symbol = self.contract_mapping[int(trade_tick[0])]
             for ticker_id in self.parent_app.symbol_mkt_id_map[symbol]:
-                # Assigning new data
                 self.parent_app.live_data[ticker_id].symbol_id = int(trade_tick[0])
-                self.parent_app.live_data[ticker_id].timestamp = datetime.strptime(trade_tick[1], '%Y-%m-%dT%H:%M:%S')  # Old format = '%m/%d/%Y %I:%M:%S %p'
+                self.parent_app.live_data[ticker_id].timestamp = datetime.strptime(trade_tick[1], '%Y-%m-%dT%H:%M:%S')
                 self.parent_app.live_data[ticker_id].symbol = symbol
                 self.parent_app.live_data[ticker_id].ltp = self.parent_app.touchline_data[ticker_id].ltp = ltp = float(trade_tick[2])
                 self.parent_app.live_data[ticker_id].ltq = float(trade_tick[3])
                 self.parent_app.live_data[ticker_id].atp = float(trade_tick[4])
                 self.parent_app.live_data[ticker_id].ttq = float(trade_tick[5])
-                self.parent_app.live_data[ticker_id].day_open = float(trade_tick[6])
+                self.parent_app.live_data[ticker_id].day_open = self.parent_app.touchline_data[ticker_id].open = float(trade_tick[6])
                 self.parent_app.live_data[ticker_id].day_high = float(trade_tick[7])
                 self.parent_app.live_data[ticker_id].day_low = float(trade_tick[8])
                 self.parent_app.live_data[ticker_id].prev_day_close = float(trade_tick[9])
@@ -183,11 +186,9 @@ class LiveClient(WebSocketApp):
                 self.parent_app.live_data[ticker_id].special_tag = special_tag = str(trade_tick[13])
                 if special_tag != "":
                     if special_tag == 'H':
-                        self.parent_app.live_data[ticker_id].day_high = self.parent_app.touchline_data[ticker_id].high = ltp
+                        self.parent_app.touchline_data[ticker_id].high = ltp
                     elif special_tag == 'L':
-                        self.parent_app.live_data[ticker_id].day_low = self.parent_app.touchline_data[ticker_id].low = ltp
-                    elif special_tag == 'O' or special_tag == 'OHL':
-                        self.parent_app.live_data[ticker_id].day_open = self.parent_app.touchline_data[ticker_id].open = ltp
+                        self.parent_app.touchline_data[ticker_id].low = ltp
                 self.parent_app.live_data[ticker_id].tick_seq = int(trade_tick[14])
                 self.parent_app.live_data[ticker_id].tick_type = 1
                 try:
@@ -205,10 +206,15 @@ class LiveClient(WebSocketApp):
                         pass
                 except Exception as e:
                     self.logger.error(e)
+                if self.trade_callback:
+                    try:
+                        self.trade_callback(ticker_id, self.parent_app.live_data[ticker_id])
+                    except Exception as e:
+                        self.logger.error(f'{Style.BRIGHT}{Fore.RED}Encountered error with trade_callback - {type(e)} - {e}{Style.RESET_ALL}')
         except KeyError:
             pass
         except Exception as e:
-            self.logger.error(f'{Style.BRIGHT}{Fore.RED}Encountered with tick feed - {type(e)}{Style.RESET_ALL}')
+            self.logger.error(f'{Style.BRIGHT}{Fore.RED}Encountered error with tick feed - {type(e)}{Style.RESET_ALL}')
 
     def handle_bid_ask_data(self, bidask_tick):
         try:
@@ -221,10 +227,15 @@ class LiveClient(WebSocketApp):
                 self.parent_app.live_data[ticker_id].best_ask_price = float(bidask_tick[4])
                 self.parent_app.live_data[ticker_id].best_ask_qty = int(bidask_tick[5])
                 self.parent_app.live_data[ticker_id].tick_type = 2
+                if self.bidask_callback:
+                    try:
+                        self.bidask_callback(ticker_id, self.parent_app.live_data[ticker_id])
+                    except Exception as e:
+                        self.logger.error(f'{Style.BRIGHT}{Fore.RED}Encountered error with bidask_callback - {type(e)} - {e}{Style.RESET_ALL}')
         except KeyError:
             pass
         except Exception as e:
-            self.logger.error(f'{Style.BRIGHT}{Fore.RED}Bid-ask feed encountered - {e}{Style.RESET_ALL}')
+            self.logger.error(f'{Style.BRIGHT}{Fore.RED}Bid-ask feed error encountered ({type(e)})- {e}{Style.RESET_ALL}')
 
     def handle_bar_data(self, bar_data):
         try:
@@ -247,6 +258,11 @@ class LiveClient(WebSocketApp):
                 obj_to_edit.close = self.parent_app.touchline_data[ticker_id].ltp = float(bar_data[5])
                 obj_to_edit.volume = float(bar_data[6])
                 obj_to_edit.oi = float(bar_data[7])
+                if self.bar_callback:
+                    try:
+                        self.bar_callback(ticker_id, obj_to_edit)
+                    except Exception as e:
+                        self.logger.error(f'{Style.BRIGHT}{Fore.RED}Encountered error with bar_callback - {type(e)} - {e}{Style.RESET_ALL}')
         except KeyError:
             pass
         except Exception as e:
@@ -258,6 +274,7 @@ class LiveClient(WebSocketApp):
             self.logger.error(f"Raising WS error = {error}")
             self.last_ping_tm = self.last_pong_tm = 0
 
+    # noinspection PyUnusedLocal
     def ws_open(self, *args):
         self.last_ping_tm = time.time()
         self.sock.ping()
@@ -266,18 +283,19 @@ class LiveClient(WebSocketApp):
             old_symbol_map = temp_symbol_map = deepcopy(self.parent_app.symbol_mkt_id_map)  # temp_symbol_map is destructed while restarting the data
             temp_symbol_map = dict(temp_symbol_map)
             self.parent_app.symbol_mkt_id_map = defaultdict(set)
-            print(f'{list(temp_symbol_map.keys())} needs resuming here.')
+            self.logger.debug(f'{list(temp_symbol_map.keys())} needs resuming here.')
             while temp_symbol_map:
                 restart_symbols = list(temp_symbol_map.keys())
                 restart_req_ids = [temp_symbol_map[symbol].pop() for symbol in restart_symbols]
                 self.parent_app.start_live_data(restart_symbols, req_id=restart_req_ids)
                 temp_symbol_map = {symbol: symbol_set for (symbol, symbol_set) in temp_symbol_map.items() if symbol_set}
             assert old_symbol_map == self.parent_app.symbol_mkt_id_map
-            print(f'All streaming symbols have been resumed.')
+            self.logger.info(f'All streaming symbols have been resumed.')
 
+    # noinspection PyUnusedLocal
     def ws_close(self, *args):
         # self.logger.error('Live WebSocket Closed')
-        self.sock.close()
+        self.sock.close() if self.sock is not None else 0
         self.sock = None
         if self.last_ping_tm == 0 == self.last_pong_tm:
             self.logger.debug('DISCONNECTION TYPE(1) FROM SERVER...')
